const AUTHORIZATION_STRING_PART = "Passport1.4 OrgVerb=GET,OrgURL=http%3A%2F%2Fmessenger%2Emsn%2Ecom,sign-in=";
const CLIENT_CAPABILITY_FLAGS = 0x10000000;
const CLIENT_ID_CODE = "Q1P7W2E4J9R8U3S5";
const CLIENT_ID_STRING = "msmsgs@msnmsgr.com";
const COMMANDS_WITHOUT_TRID = ["QNG", "NLN", "FLN", "MSG", "GTC", "BLP", "LST",
                               "CHL", "NOT"];
const SESSION_ERRORS = [205, 208, 209, 210, 215, 800];
const LOG_IN_ERRORS = [500, 540, 601, 910, 911, 921];
const DISPATCH_SERVER_URL = "messenger.hotmail.com";
const PASSPORT_NEXUS_URL = "https://nexus.passport.com/rdr/pprdr.asp";

var MsnProtocol = {
  port : "1863",
  versions : ["MSNP9", "MSNP8"]
};

MsnProtocol.onConnectionClosed = function(aStatus) {
  if (aStatus) {
    this.msnSession.onConnectionClosed(aStatus);
  }
}

MsnProtocol.onErrorReceived = function(aErrorCode) {
  if (aErrorCode in SESSION_ERRORS) {
    this.msnSession.onErrorReceived(aErrorCode);
  }
  else if (aErrorCode in LOG_IN_ERRORS) {
    this.msnSession.onLogInFailed(aErrorCode);
  }
  else if (aErrorCode == 913) {
    delete this.newSession;
    this.msnSession.onErrorReceived(913);
  }
}

MsnProtocol.onHttpDataAvailable = function(aStatus) {
  if (this.authPhase == 1) {
    var passportUrls = this.httpSocket.getHeader("PassportURLs");
    var loginServerUrl = "https://" + extractField(passportUrls, "DALogin");
    dump("Login Server URL received. Opening connection to Login Server...\n");
    delete this.httpSocket;
    this.authPhase = 2;
    this.httpSocket = new HttpSocket(loginServerUrl);
    var authorizationHeader = AUTHORIZATION_STRING_PART
     + encodeURIComponent(this.msnSession.passport)
     + ",pwd="
     + encodeURIComponent(this.msnSession.password)
     + ","
     + this.challenge;
    this.httpSocket.setHeader("Authorization", authorizationHeader);
    this.httpSocket.sendRequest(this);
    delete this.challenge;
  }
  else {
    var authenticationInfo = this.httpSocket.getHeader("Authentication-Info");
    var ticket = extractField(authenticationInfo, "from-PP");
    ticket = ticket.substr(1, ticket.length - 2);
    this.msnSession.mspAuth = ticket.substr(2, ticket.search("&p") - 2);
    delete this.httpSocket;
    delete this.authPhase;
    dump("Ticket received. Sending ticket to Notification Server...\n");
    this.connection.sendMsg(["USR", ++this.trId, "TWN S", ticket]);
    this.msnSession.onLogInProgress(3);
  }
}

MsnProtocol.onProtocolMessageReceived = function(aMsg, aPayload) {
  if (COMMANDS_WITHOUT_TRID.indexOf(aMsg[0]) == -1 && aMsg[1] == "0"
      && aMsg[0] != "ILN" && aMsg[0] != "ADD" && aMsg[0] != "REM") {
    dump("Command ignored.\n");
    return;
  }
  if (aMsg[0] == "MSG") {
    this.msgReceived(aMsg, aPayload);
  }
  else {
    var functionToCall = aMsg[0].toLowerCase() + "Received";
    if (functionToCall in this) {
      this[functionToCall](aMsg);
    }
    else if (isNumeric(aMsg[0])) {
      var errorCode = parseInt(aMsg[0]);
      this.onErrorReceived(errorCode);
    }
  }
}

MsnProtocol.logIn = function() {
  this.connection = new BaseConnection(this);
  this.trId = 0;
  this.connection.connect(DISPATCH_SERVER_URL, this.port);
  this.connection.sendMsg(["VER", ++this.trId, this.versions.join(" ")]);
  this.msnSession.onLogInProgress(0);
}

MsnProtocol.logOut = function() {
  if (this.connection) {
    if (this.connection.isConnected) {
      this.connection.sendMsg(["OUT"]);
      this.connection.disconnect();
    }
    delete this.connection;
  }
  if (this.httpSocket) {
    this.httpSocket.close();
    delete this.httpSocket;
  }
  clearTimeout(this.pinger);
  clearInterval(this.sessionLengthUpdater);
}

MsnProtocol.addContactToList = function(aEmail, aDisplayName, aListName) {
  var displayName = encodeURIComponent(aDisplayName);
  if (aListName == "FL") {
    this.sendCommand(["ADD", ++this.trId, aListName, aEmail, displayName, "0"]);
  }
  else {
    this.sendCommand(["ADD", ++this.trId, aListName, aEmail, displayName]);
  }
}

MsnProtocol.changeDisplayName = function(aDisplayName) {
  var displayName = encodeURIComponent(aDisplayName);
  this.sendCommand(["REA", ++this.trId, this.msnSession.passport, displayName]);
}

MsnProtocol.changeStatus = function(aStatus, aMsnObjectString) {
  if (aMsnObjectString && this.version == this.versions[0]) {
    var msnObjectString = encodeURIComponent(aMsnObjectString);
    this.sendCommand(["CHG", ++this.trId, aStatus, CLIENT_CAPABILITY_FLAGS,
                      msnObjectString]);
  }
  else {
    this.sendCommand(["CHG", ++this.trId, aStatus, CLIENT_CAPABILITY_FLAGS]);
  }
}

MsnProtocol.ping = function() {
  this.connection.sendMsg(["PNG"]);
}

MsnProtocol.removeContactFromList = function(aEmail, aListName) {
  if (aListName == "FL") {
    this.sendCommand(["REM", ++this.trId, aListName, aEmail, "0"]);
  }
  else {
    this.sendCommand(["REM", ++this.trId, aListName, aEmail]);
  }
}

MsnProtocol.requestEmailInboxLogIn = function() {
  this.connection.sendMsg(["URL", ++this.trId, "INBOX"]);
}

MsnProtocol.requestEmailMessageComposeLogIn = function(aEmail) {
  this.connection.sendMsg(["URL", ++this.trId, "COMPOSE"]);
}

MsnProtocol.requestSbSession = function(sbSession) {
  this.newSbSession = sbSession;
  this.sendCommand(["XFR", ++this.trId, "SB"]);
}

MsnProtocol.sendCommand = function(aCommand) {
  this.connection.sendMsg(aCommand);
  clearTimeout(this.pinger);
  this.pinger = setTimeout(function() { MsnProtocol.ping(); }, 50000);
}

MsnProtocol.setMsnSession = function(aMsnSession) {
  this.msnSession = aMsnSession;
}

MsnProtocol.updateContactDisplayName = function(aEmail, aDisplayName) {
  var displayName = encodeURIComponent(aDisplayName);
  this.sendCommand(["REA", ++this.trId, aEmail, displayName]);
}

MsnProtocol.addReceived = function(aMsg) {
  // ADD trId listName contactListVersion email displayName
  this.msnSession.onContactAddedToList(aMsg[4], aMsg[2]);
}

MsnProtocol.blpReceived = function(aMsg) {
  // BLP blpSetting
  this.msnSession.blp = aMsg[1];
}

MsnProtocol.chgReceived = function(aMsg) {
  // CHG trId statusCode capabilityFlags 
  var statusCode = aMsg[2];
  this.msnSession.onStatusChanged(statusCode);
}

MsnProtocol.chlReceived = function(aMsg) {
  // CHL zero challenge
  var md5Hash = Utils.getMd5HashFromStringAsHex(aMsg[2] + CLIENT_ID_CODE);
  this.connection.sendMsg(["QRY", ++this.trId, CLIENT_ID_STRING], md5Hash);
  clearTimeout(this.pinger);
  this.pinger = setTimeout(function() { MsnProtocol.ping(); }, 50000);
}

MsnProtocol.cvrReceived = function(aMsg) {
  // CVR trId recommendedVersion recommendedVersion minimumVersion
  //     recommendedVersionDownloadUrl moreInfoUrl
  this.connection.sendMsg(["USR", ++this.trId, "TWN I",
                           this.msnSession.passport]);
}

MsnProtocol.flnReceived = function(aMsg) {
  // FLN email
  this.msnSession.onContactSignedOut(aMsg[1]);
}

MsnProtocol.gtcReceived = function(aMsg) {
  // GTC gtcSetting
  this.msnSession.gtc = aMsg[1];
}

MsnProtocol.ilnReceived = function(aMsg) {
  // ILN trId statusCode email displayName capabilityFlags
  var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[4]));
  var capabilityFlags = parseInt(aMsg[5]);
  var msnObjectString = (aMsg.length > 6) ? decodeURIComponent(aMsg[6]) : "";
  this.msnSession.onContactAlreadySignedIn(aMsg[3], displayName, aMsg[2],
                                           capabilityFlags, msnObjectString);
}

MsnProtocol.lstReceived = function(aMsg) {
  // LST email displayName listFlags group1,group2,...
  var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[2]));
  var listFlags = parseInt(aMsg[3]);
  this.msnSession.onContactReceived(aMsg[1], displayName, listFlags);
  this.contactsLeft--;
  if (this.contactsLeft == 0) {
    delete this.contactsLeft;
    this.msnSession.onContactListReceived();
  }
}

MsnProtocol.msgReceived = function(aMsg, aPayload) {
  // MSG email displayName payloadLength
  var msgParts = aPayload.split("\r\n\r\n");
  var headers = Utils.parseHeaders(msgParts[0]);
  switch (headers["content-type"]) {
    case "text/x-msmsgsprofile; charset=UTF-8":
      this.msnSession.ipAddress = headers["clientip"];
      this.ping();
      this.msnSession.onLogInSuccessful();
      break;
    case "text/x-msmsgsinitialemailnotification; charset=UTF-8":
      var msgData = Utils.parseHeaders(msgParts[1]);
      var nUnreadEmails = parseInt(msgData["inbox-unread"]);
      this.msnSession.onInitialEmailNotification(nUnreadEmails);
      break;
    case "text/x-msmsgsemailnotification; charset=UTF-8":
      var msgData = Utils.parseHeaders(msgParts[1]);
      var sender = msgData["from"];
      var arrivedInInbox = (msgData["dest-folder"] == "ACTIVE");
      this.msnSession.onNewEmailMessageReceived(sender, arrivedInInbox);
      break;
    case "text/x-msmsgsactivemailnotification; charset=UTF-8":
      var msgData = Utils.parseHeaders(msgParts[1]);
      var srcFolder = msgData["src-folder"];
      var destFolder = msgData["dest-folder"];
      var nEmails = parseInt(msgData["message-delta"]);
      if (srcFolder == "ACTIVE" && destFolder == "ACTIVE") {
        this.msnSession.onEmailMessagesRead(nEmails);
      }
      else if (srcFolder == "ACTIVE") {
        this.msnSession.onEmailMessagesRemoved(nEmails);
      }
      else if (destFolder == "ACTIVE") {
        this.msnSession.onEmailMessagesRecovered(nEmails);
      }
      break;
  }
}

MsnProtocol.nlnReceived = function(aMsg) {
  // NLN statusCode email displayName capabilityFlags
  var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[3]));
  var capabilityFlags = parseInt(aMsg[4]);
  var msnObjectString = (aMsg.length > 5) ? decodeURIComponent(aMsg[5]) : "";
  this.msnSession.onContactChanged(aMsg[2], displayName, aMsg[1], capabilityFlags,
                                   msnObjectString);
}

MsnProtocol.qngReceived = function(aMsg) {
  // QNG
  clearTimeout(this.pinger);
  if (aMsg.length == 2) {
    var timeout = parseInt(aMsg[1]) * 1000;
    this.pinger = setTimeout(function() { MsnProtocol.ping(); }, timeout);
  }
  else {
    var timeout = 50 * 1000;
    this.pinger = setTimeout(function() { MsnProtocol.ping(); }, timeout);
  }
}

MsnProtocol.reaReceived = function(aMsg) {
  // REA trId contactListVersion email displayName
  if (aMsg[3] == this.msnSession.passport) {
    var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[4]));
    this.msnSession.onDisplayNameChanged(displayName);
  }
}

MsnProtocol.remReceived = function(aMsg) {
  // REM trId listName contactListVersion email groupId
  this.msnSession.onContactRemovedFromList(aMsg[4], aMsg[2]);
}

MsnProtocol.rngReceived = function(aMsg) {
  // RNG sessionId sbIp:sbPort authType authToken email displayName
  var sbAddress = aMsg[2].split(":");
  var email = aMsg[5];
  var sb = {
    ip : sbAddress[0],
    port : sbAddress[1],
    cki : aMsg[4],
    sessionId : aMsg[1]
  };
  this.msnSession.openSbSession(email, sb);
}

MsnProtocol.synReceived = function(aMsg) {
  // SYN trId contactListVersion nContact nGroups
  if (aMsg[3] != "0") {
    this.contactsLeft = parseInt(aMsg[3]);
  }
  else {
    this.changeStatus(this.msnSession.status);
    this.msnSession.onLogInSuccessful();
  }
}

MsnProtocol.urlReceived = function(aMsg) {
  // URL trId rru action id
  this.msnSession.onEmailLogInReceived(aMsg[2], aMsg[3], aMsg[4]);
}

MsnProtocol.usrReceived = function(aMsg) {
  // USR trId authMethod S authToken
  // USR OK email displayName passportVerified unknownNumber
  if (aMsg[3] == "S") {
    this.challenge = aMsg[4];
    this.msnSession.sid = extractField(this.challenge, "id");
    this.msnSession.kv = extractField(this.challenge, "kv");
    dump("Opening connection to Passport Nexus...\n");
    this.authPhase = 1;
    this.httpSocket = new HttpSocket(PASSPORT_NEXUS_URL);
    this.httpSocket.sendRequest(this);
    this.msnSession.onLogInProgress(2);
  }
  else if (aMsg[2] == "OK") {
    displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[4]));
    this.msnSession.onDisplayNameChanged(displayName);
    this.connection.sendMsg(["SYN", ++this.trId, "0"]);
    this.msnSession.onLogInProgress(4);
  }
}

MsnProtocol.verReceived = function(aMsg) {
  // VER trId preferredVersion1 preferredVersion2 ...
  if (aMsg[2] != "0") {
    this.version = aMsg[2];
    var localeId = this.msnSession.localeId;
    var platform = this.msnSession.platform;
    this.connection.sendMsg(["CVR", ++this.trId, localeId, platform,
                             "MSNMSGR 5.0.0575 MSMSGS",
                             this.msnSession.passport]);
  }
}

MsnProtocol.xfrReceived = function(aMsg) {
  // XFR trId serverType ip:port zero currentServerIp:currentServerPort
  var host = aMsg[3].split(":");
  if (aMsg[2] == "NS") {
    this.connection.disconnect();
    this.trId = 0;
    this.connection.connect(host[0], host[1]);
    this.connection.sendMsg(["VER", ++this.trId, this.versions.join(" ")]);
    this.msnSession.onLogInProgress(1);
  }
  else if (aMsg[2] == "SB") {
    this.newSbSession.open(host[0], host[1], aMsg[5]);
    delete this.newSbSession;
  }
}

function extractField(aFields, aFieldKey) {
  var field = aFields.substr(aFields.search(aFieldKey));
  return field.substr(aFieldKey.length + 1, field.indexOf(",")
                      - (aFieldKey.length + 1));
}

function isNumeric(aString) {
  var numbers = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "0"];
  for (var character in aString) {
    if (!(character in numbers)) {
      return false;
    }
  }
  return true;
}