function P2PSession(aSessionId, aContactId, aParent) {
  this.sessionId = aSessionId;
  this.contactId = aContactId;
  this.parent = aParent;
}

P2PSession.prototype.on200MessageReceived = function(aHeaders, aContent) {
  if (aHeaders["content-type"] == "application/x-msnmsgr-sessionreqbody"
      && this.appId == FILE_TRANSFER_APP_ID) {
    this.sendFile();
  }
}

P2PSession.prototype.onAcknowledgementReceived = function() {
  dump("Acknowledgement received!\n");
  if (this.sendNextMessage) {
    this.sendNextMessage();
  }
}

P2PSession.prototype.onSessionMessageReceived = function(aHeaders, aMsg) {
  if (aMsg.length == 0) {
    dump("Empty message with flags " + aHeaders.flags + " received!\n");
    if (aHeaders.flags == FLAG_ACKNOWLEDGEMENT) {
      this.onAcknowledgementReceived();
    }
  }
  else if (aMsg.length == 4 & aMsg == "\0\0\0\0") {
    dump("Data preparation message received!\n");
    this.fileOutputStream = Components.classes["@mozilla.org/network/safe-file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
    this.fileOutputStream.init(this.file, 0x04 | 0x08 | 0x20, 0600, 0);
  }
  else {
    if (!this.fileOutputStream) {
      this.fileOutputStream = Components.classes["@mozilla.org/network/safe-file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
      this.fileOutputStream.init(this.file, 0x04 | 0x08 | 0x20, 0600, 0);
      if (this.appId == FILE_TRANSFER_APP_ID) {
        this.parent.showFileTransferStatus("receiving");
      }
    }
    this.fileOutputStream.write(aMsg, aMsg.length);
    if (this.appId == FILE_TRANSFER_APP_ID) {
      var progressPercentage = Math.floor(aHeaders.dataOffset / aHeaders.msgLength * 100);
      this.parent.showFileTransferProgress(progressPercentage);
    }
    if (aHeaders.dataOffset + aMsg.length == aHeaders.msgLength) {
      if (this.fileOutputStream instanceof Components.interfaces.nsISafeOutputStream) {
        this.fileOutputStream.finish();
      }
      else {
        this.fileOutputStream.close();
      }
      delete this.fileOutputStream;
      this.parent.sendByeRequest(this.contactId, this.callId);
      if (this.appId == DISPLAY_PICTURE_APP_ID_1
          || this.appId == DISPLAY_PICTURE_APP_ID_2) {
        var fileUri = ioService.newFileURI(this.file);
        this.parent.onDisplayPictureReceived(fileUri.spec);
      }
      else if (this.appId == FILE_TRANSFER_APP_ID) {
        this.parent.onFileTransferred(this.file.leafName);
      }
      this.parent.onSessionClosed(this.sessionId);
    }
  }
}

P2PSession.prototype.onInviteRequestReceived = function(aHeaders, aContent) {
  if (aHeaders["content-type"] == "application/x-msnmsgr-sessionreqbody") {
    this.callId = aHeaders["call-id"];
    var viaParts = aHeaders["via"].split("=");
    var branchUid = viaParts[1];
    if (aContent["euf-guid"] == MSN_OBJECT_GUID) {
      this.appId = DISPLAY_PICTURE_APP_ID_1;
      this.file = Utils.getExtensionSubDirectory("pictures");
      this.file.append(btoa(this.parent.passport));
      this.sendNextMessage = this.sendDataPreparationMessage;
      this.parent.send200Message(this.contactId, branchUid, aHeaders["call-id"], this.sessionId);
    }
    else if (aContent["euf-guid"] == FILE_TRANSFER_GUID) {
      this.appId = FILE_TRANSFER_APP_ID;
      var context = atob(aContent["context"]);
      var contextObj = this.parent.getDecodedContext(context);
      var file = this.parent.acceptFile(contextObj.fileName, contextObj.fileSize);
      if (file) {
        this.file = file;
        this.parent.send200Message(this.contactId, branchUid, aHeaders["call-id"], this.sessionId);
        this.parent.showFileTransferStatus("waiting");
      }
      else {
        this.parent.send603Message(this.contactId, branchUid, aHeaders["call-id"], this.sessionId);
        this.parent.onSessionClosed(this.sessionId);
      }
    }
  }
}

P2PSession.prototype.inviteForFileTransfer = function(aFile) {
  this.appId = FILE_TRANSFER_APP_ID;
  this.callId = Utils.getUniversallyUniqueId();
  this.file = aFile;
  var context = this.parent.getEncodedContext(this.file.leafName, this.file.fileSize);
  this.parent.sendInviteRequest(this.contactId, this.callId, FILE_TRANSFER_GUID, this.sessionId, this.appId, context);
}

P2PSession.prototype.requestDisplayPicture = function(aMsnObj) {
  this.appId = DISPLAY_PICTURE_APP_ID_1;
  this.callId = Utils.getUniversallyUniqueId();
  var file = Utils.getExtensionSubDirectory("pictures");
  file.append(btoa(this.contactId));
  this.file = file;
  this.parent.sendInviteRequest(this.contactId, this.callId, MSN_OBJECT_GUID, this.sessionId, this.appId, aMsnObj);
}

P2PSession.prototype.sendDataPreparationMessage = function() {
  this.sendNextMessage = this.sendFile;
  this.parent.sendDataPreparationMessage(this.sessionId, this.contactId, this.appId);
}

P2PSession.prototype.sendFile = function() {
  this.sendNextMessage = null;
  this.fileInputStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
  this.binaryInputStream = Components.classes["@mozilla.org/binaryinputstream;1"].createInstance(Components.interfaces.nsIBinaryInputStream);
  this.fileInputStream.init(this.file, -1, -1, false);
  this.binaryInputStream.setInputStream(this.fileInputStream);
  if (this.appId == FILE_TRANSFER_APP_ID) {
    this.parent.showFileTransferStatus("sending");
  }
  var nBytesLeft = this.file.fileSize;
  while (nBytesLeft > 0) {
    var nBytesToRead = (nBytesLeft >= 1200) ? 1200 : nBytesLeft;
    var bytes = this.binaryInputStream.readBytes(nBytesToRead);
    var dataOffset = this.file.fileSize - nBytesLeft;
    this.parent.sendDataMessage(this.sessionId, this.contactId, dataOffset, this.file.fileSize, nBytesToRead, this.appId, bytes);
    nBytesLeft -= nBytesToRead;
    if (this.appId == FILE_TRANSFER_APP_ID) {
      var progressPercentage = Math.floor(((this.file.fileSize - nBytesLeft) / this.file.fileSize) * 100);
      this.parent.showFileTransferProgress(progressPercentage);
    }
  }
  delete this.binaryInputStream;
  this.fileInputStream.close();
  delete this.fileInputStream;
  if (this.appId == FILE_TRANSFER_APP_ID) {
    this.parent.onFileTransferred(this.file.leafName);
  }
}